import regex as re
import numpy as np
from difflib import SequenceMatcher

pattern = re.compile(r"""'(?i:[sdmt]|ll|ve|re)|[^\r\n\p{L}\p{N}]?+\p{L}+|\p{N}{1,3}| ?[^\s\p{L}\p{N}]++[\r\n]*|\s*[\r\n]|\s+(?!\S)|\s+""")  # GPT4 pre tokenize


def calculate_cer(s1, s2):
    """
    Character Error Rate (CER) is typically (substitutions + deletions + insertions) / len(original)
    We'll use a simplistic version where we just count how many characters differ
    This uses SequenceMatcher which gives a similarity ratio; we convert it to error rate
    """
    matcher = SequenceMatcher(None, s1, s2)
    return 1 - matcher.ratio()


def calculate_wer(reference, hypothesis):
    """
    Calculates the Word Error Rate (WER) between a reference and a hypothesis sentence.
    """
    if reference == hypothesis == "":
        return 0

    elif reference == "" or hypothesis == "":
        return 1

    reference_words = reference.split()
    hypothesis_words = hypothesis.split()
    matrix = np.zeros((len(reference_words) + 1, len(hypothesis_words) + 1))

    for i in range(len(reference_words) + 1):
        matrix[i, 0] = i
    for j in range(len(hypothesis_words) + 1):
        matrix[0, j] = j
    for i in range(1, len(reference_words) + 1):
        for j in range(1, len(hypothesis_words) + 1):
            if reference_words[i - 1] == hypothesis_words[j - 1]:
                matrix[i, j] = matrix[i - 1, j - 1]
            else:
                substitution = matrix[i - 1, j - 1] + 1
                insertion = matrix[i, j - 1] + 1
                deletion = matrix[i - 1, j] + 1
                matrix[i, j] = min(substitution, insertion, deletion)

    distance = matrix[len(reference_words), len(hypothesis_words)]
    wer_result = distance / len(reference_words)
    return wer_result


def calculate_cer_full(reference: str, hypothesis: str):
    """
    Calculate the Character Error Rate (CER) between reference and hypothesis strings.
    """
    ref = reference.replace(" ", "")
    hyp = hypothesis.replace(" ", "")
    n, m = len(ref), len(hyp)

    d = np.zeros((n+1)*(m+1), dtype=np.uint8).reshape((n+1, m+1))
    for i in range(n+1):
        for j in range(m+1):
            if i == 0:
                d[0][j] = j
            elif j == 0:
                d[i][0] = i

    for i in range(1, n+1):
        for j in range(1, m+1):
            if ref[i-1] == hyp[j-1]:
                cost = 0
            else:
                cost = 1
            d[i][j] = min(d[i-1][j] + 1,      # Deletion
                          d[i][j-1] + 1,      # Insertion
                          d[i-1][j-1] + cost) # Substitution

    cer = d[n][m] / float(n)
    return cer



from typing import Optional
from dataclasses import dataclass


@dataclass(frozen=True)
class ChunkData:
    save_path: str
    audio_path: str
    duration: float
    asr_prediction: str
    text: Optional[str] = None


@dataclass(frozen=True)
class ASRConfig:
    """
    Configuration for Automatic Speech Recognition (ASR).

    Attributes:
        hf_token (str): Hugging Face API token for authentication.
        gpu_id (Optional[int]): ID of the GPU to use (None if no GPU).
        batch_size (int): The number of samples to process in a batch.
    """
    hf_token: str
    batch_size: int = 15
    gpu_id: Optional[int] = 0


@dataclass(frozen=True)
class ALIGNERConfig:
    """
    Configuration for the Aligner.

    Attributes:
        manifest_file (str):
            The path to the NEMO manifest file. Each line in this file should be a valid JSON string representing
            individual audio data with keys [audio_filepath, pred_text, duration].

        transcript_path (Optional[str]):
            The path to the file containing the combined transcript of all audio files.
            If None, it will be calculated (combined) based on the "text" key in the manifest_file.

        save_manifest_path (str):
            The path where the processed manifest data will be saved or utilized.

        use_id (bool):
            Determines whether to sort the audio chunks based on their ID. Useful when IDs are used to denote the
            order of audio files (chunks). Default is False.

        target_base (Optional[str]):
            The base path that could be prefixed to each audio file path in the manifest, useful for cases where
            the manifest paths are relative. Default is None.

        ending_punctuations (Optional[str]):
            The punctuation symbols accumulated in a single string (with the most common sentence-ending punctuation
            placed at the end of the string, such as "...,;:."). Only necessary if transcript_path is None, so that
            the texts can be combined into a single transcript. Default is an example for Armenian Language - '․,։'.
    """
    manifest_file: str
    save_manifest_path: str

    use_id: bool = False
    target_base: Optional[str] = None
    transcript_path: Optional[str] = None
    ending_punctuations: Optional[str] = '․,։'



import pytest
import warnings

from vac_aligner.matching.mapping import cut_extra_tokens_from_match

warnings.filterwarnings('ignore')


@pytest.mark.parametrize("reference_text, asr_bad_pred, expected", [
    (
    "EXTRA1 EXTRA2, Text to Be Mached EXTRA3 but mistakenly shifted too much", "Txt  tobe, mtchd", 'Text to Be Mached'),
    ("I amm old boy: mistakenly shifted window too much", "I am old boy", 'amm old boy:'),
    ('Իր նիզակը ներսուք տալով մարմնին, կանգնեց պատասխանելով։',
     'իզակը նեցուկ տալով մարմնին, կանգնեց, պատասխանելով. Վանից։',
     'նիզակը ներսուք տալով մարմնին, կանգնեց պատասխանելով։'),
    ('ավաղ, տեսարանն երկար չտևեց։ Կինը համր քայլերով հեռա', 'Կինը համր քայլելով հեռացավ լուսամուտից։',
     'Կինը համր քայլերով հեռա')
])


def test_cut_extra_tokens_from_match(reference_text, asr_bad_pred, expected):
    assert cut_extra_tokens_from_match(reference_text, asr_bad_pred, try_combinations=True, verbose=False)[
               0] == expected




import regex as re

from .utils import lowercase
from .metrics import calculate_cer_full


def cut_extra_tokens_from_match(original: str, predicted: str, try_combinations: bool = False, verbose: bool = False):
    """
    Removes redundant parts from the Source Segment to find the best match with the ASR-predicted chunk text.
    This function can optionally try different combinations (cut from both start and the end) to find the best cut-offs
    in the predicted string that matches the original.

    Parameters
    ----------
    original : str
        The original (long) string to compare against - Source Segment Text from the long transcript
    predicted : str
        The predicted chunk text which is potentially full of ASR errors and needs to be replaced from the corresponding
        part from the source label (original)
    try_combinations : bool, optional
        If True, the function will attempt various combinations simultaneously to determine the
        best point to cut the predicted string. Defaults to False which means, only will cut from the end (right)
    verbose : bool, optional
        If True, the function will print detailed information about the operations
        being performed. Defaults to False.

    Returns
    -------
    str
        The predicted string after removing the redundant parts.

    Examples
    --------
    >>> original = "Arfifacts from left; This is a Source Segment that needs to be matched with the ASR preds (chunks)."
    >>> predicted =                     "This is an source segment that leads to bemach with the a s r brides."
    >>> cut_extra_tokens_from_match(original, predicted)
    'This is a Source Segment that needs to be matched with the ASR preds'  # the best possible match
    The latter is most probably (see the benchmark with 97% accuracy) the true text spoken in the chunk !
    """

    if verbose:
        print(f"Predicted chunk text (subject to replaced with corresponding source text segment): {predicted}")
        print("Note, this is much smaller than the Source Segment below, to ensure, it contains the spoken chunk text.")
        print(f"Source Text Segment  (subject to be searched over): {original}")
        print()

    to_shift = 0  # index from the left
    best_cut = original  # currently the full "lengthy" source text
    base_cer = calculate_cer_full(original, predicted)
    best_cer = base_cer

    tokens = re.findall(r'\S+|\s+', original)  # Tokenizes the string including whitespace
    tokens = list(map(lowercase, tokens))
    if not try_combinations:
        for i in range(len(tokens), 0, -1):
            cutted = "".join(tokens[:i])
            curr_cer = calculate_cer_full(cutted, predicted)
            if curr_cer <= best_cer:  # Consider '<=' to handle the case where shorter substrings might also have the same CER
                best_cer = curr_cer
                best_cut = cutted
        if verbose:
            print(f"Cut only from the right (end): {best_cut}")
    else:
        for st_idx in range(1, min(len(tokens) - 4, 7)):  # First for is to remove extra tokens from left
            for i in range(len(tokens), 0, -1):  # second for is to remove extra tokens from the right
                cutted = "".join(tokens[st_idx:i])
                curr_cer = calculate_cer_full(cutted, predicted)
                if curr_cer <= best_cer:  # Consider '<=' to handle the case where shorter substrings might also have the same CER
                    best_cer = curr_cer
                    best_cut = cutted
                    to_shift = len("".join(tokens[:st_idx]))
                    if verbose:
                        print(
                            f"Inner Loop. Cutted from the start (left): {tokens[:st_idx]}; cutting from the right (end); best cut so far: {tokens[i:]}; CER: {round(best_cer, 2)}")

            if verbose:
                print("------------Outer Loop------------")
                print(f"to_shift (from left) {to_shift}; best cut so far: {best_cut}; CER: {round(best_cer, 2)}")

        init_shift = to_shift
        tokens = re.findall(r'\S+|\s+', best_cut)
        cut_counter = 0
        for i in range(0, len(tokens)):
            cutted = "".join(tokens[i:])
            if cut_counter >= 3 and len(cutted) < len(predicted):
                break

            curr_cer = calculate_cer_full(cutted, predicted)
            if curr_cer <= best_cer:  # Consider '<=' to handle the case where shorter substrings might also have the same CER
                cut_counter = 0
                best_cer = curr_cer
                best_cut = cutted
                to_shift = init_shift + len("".join(tokens[:i]))
                if verbose:
                    print(
                        f"cut_counter: {cut_counter} to_shift: {to_shift}; init_shift: {init_shift}; best cut: {best_cut}; CER: {round(best_cer, 2)}")
            else:
                cut_counter += 1
    if verbose:
        print()
        print(f"Matched Pairs: {(predicted, best_cut)}")  # First text is not subject to change, as well as the
        # Source text, just need the best possible match
    return best_cut, best_cer, to_shift


def demo(original, predicted):
    new_b, new_cer, to_shift = cut_extra_tokens_from_match(original, predicted, try_combinations=True)
    print()
    print("Original CER: ",  round(calculate_cer_full(original, predicted), 2))
    print("Source         text:", original)
    print("Predicted      text:", predicted)
    print("Optimized substring:", new_b)
    print("Minimized CER:", round(new_cer, 2))
    print()


if __name__ == '__main__':
    import warnings

    warnings.filterwarnings('ignore')

    reference_text = "EXTRA1 EXTRA2, Text to Be Mached EXTRA3 but mistakenly shifted too much"
    asr_bad_pred = "Txt  tobe, mtchd"

    match_ = cut_extra_tokens_from_match(reference_text, asr_bad_pred, try_combinations=True, verbose=True)

    # ------------------
    demo_texts = [
        ("I amm old boy: mistakenly shifted window too much", "I am old boy"),
        ('Իր նիզակը ներսուք տալով մարմնին, կանգնեց պատասխանելով։',
         'իզակը նեցուկ տալով մարմնին, կանգնեց, պատասխանելով. Վանից։'),
        ('Իր նիզակը ներսուք տալով մարմնին, կանգնեց պատասխանելով։',
         'իզակը նեցուկ տալով մարմնին, կանգնեց, պատասխանելով. Վանից։'),
        ('ավաղ, տեսարանն երկար չտևեց։ Կինը համր քայլերով հեռա', 'Կինը համր քայլելով հեռացավ լուսամուտից։'),
        (' ՍԱՂԱԹԵԼ – Բան', 'Սաղացթել։'), (' ԿԱՐԻՆՅԱՆ – Մնա', 'Կարինյան։')
    ]

    for ex in demo_texts:
        reference_text, asr_errorours_pred = ex
        demo(reference_text, asr_errorours_pred)


import json
import argparse
from typing import List

import numpy as np
import pandas as pd
from glob import glob

from .metrics import calculate_wer
from .align import ArmenianAlignerVAC


class Benchmark:
    def __init__(self, target_base: str, predictions_manifest: str):
        """A class performing benchmark for the VAC"""
        self.predictions_manifest = predictions_manifest
        self.matched_texts: List[str] = glob(target_base+"/*.txt")

    def get_benchmark(self):
        columns = ["Text", "ASR Text", "Matched Text", "VAC_WER", "ASR_WER", "Duration", "Path"]
        benchmark = pd.DataFrame(np.zeros([len(self.matched_texts), len(columns)]), columns=columns)

        i = -1
        with open(fr"{self.predictions_manifest}", "r", encoding='utf-8') as f:
            for line in f:
                i += 1
                item = json.loads(line)
                orig_text = item["text"]
                duration = item["duration"]
                path = item["audio_filepath"]
                if "\\" in path:
                    reference = path.split("\\")[-1].replace(".wav", "_matched.txt")
                else:
                    reference = path.split("/")[-1].replace(".wav", "_matched.txt")
                matched_text_file = [x for x in self.matched_texts if reference in x]
                assert len(matched_text_file) == 1, f"matched_text_file: {matched_text_file}"
                matched_text_file = matched_text_file[0]
                with open(matched_text_file, "r", encoding='utf-8') as f:
                    matched_text = f.read().strip()

                pred_text = item['pred_text']
                asr_wer = calculate_wer(orig_text,pred_text)
                vac_wer = calculate_wer(orig_text, matched_text)
                benchmark.iloc[i, :] = [orig_text, pred_text, matched_text, vac_wer, item.get('wer', asr_wer), duration, path]
        return benchmark

    @staticmethod
    def analyze_and_save_benchmark(benchmark: pd.DataFrame, output_file: str):
        benchmark.VAC_WER = benchmark.VAC_WER.apply(lambda x: 0. if x == 0 else round(x, 3))
        benchmark.ASR_WER = benchmark.ASR_WER.apply(lambda x: 0. if x == 0 else round(x, 3))
        print(f"Mean WER: {round(100 * benchmark.VAC_WER.sum() / len(benchmark), 3)}%")
        print(f"Accuracy (ASR): {round(100 * len(benchmark[benchmark.ASR_WER == 0]) / len(benchmark.ASR_WER), 3)}%")
        print(f"Accuracy (VAC): {round(100 * len(benchmark[benchmark.VAC_WER == 0]) / len(benchmark.VAC_WER), 3)}%")
        benchmark.to_csv(output_file.replace(".txt", ".csv"), index=False)
        print(benchmark.sample(10))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Process the predictions manifest and output a combined transcript.")
    parser.add_argument("--output_file", type=str, help="Path where to save the combined MCV transcript.", default=r"D:\CapstoneThesisArmASR\publish\experiments\combined.txt")
    parser.add_argument("--target_base", type=str, help="Path where need to save matched texts", default=r"D:\CapstoneThesisArmASR\publish\experiments\matches")
    parser.add_argument("--predictions_manifest", type=str, help="Path to the manifest file storing predictions.",
                        default=r"C:\Users\GIGABYTE\workspace\Nvidia\NeMo\test_with_predictions17_replaced.json")
    parser.add_argument("--save_manifest_path", type=str, help="Path to save matching related manifest", default="D:\CapstoneThesisArmASR\publish\experiments\matches.json")
    args = parser.parse_args()

    chunks, combined_transcript = ArmenianAlignerVAC.combine_transcript(args.predictions_manifest,
                                                                        args.output_file, ending_punctuations="․,։")
    print("Combined the MCV files into single transcript")
    # AlignerVAC.align(combined_transcript, chunks, args.output_file, target_base=args.target_base)
    matches_sorted = ArmenianAlignerVAC(combined_transcript, chunks, args.save_manifest_path, target_base=args.target_base).align(0.35)

    """ Logs You should get (timestamp can vary from system to system)
    1it [00:00,  1.90it/s]~/AppData/Local/Temp/ipykernel_33732/804255406.py:31: RuntimeWarning: divide by zero encountered in divide
      cer = d[n][m] / float(n)
    48it [00:17,  2.64it/s]['Ինչպես հաղորդվում է՝ ֆիլմը կատակերգությու', 'Ինչպես հավաք է, մկատագիտությունը, ։', 0.5405405405405406, 3152, 3193, 46]
    49it [00:20,  1.20s/it]['հաղորդվում է՝ ֆիլմը կատակերգություն է լինելու։ Ռուսաստանում այն ստացավ պլատինե', 'Սոեբում նաի ստացավ լա, եստտավե ան համ, ավաստան։', 0.7391304347826086, 3159, 3237, 47]
    137it [00:45,  6.89it/s][' Է՛հ, գնա՛', ' է՛՜հ, գնա է։', 0.5, 8152, 8162, 136]
    609it [02:55,  2.21it/s]['րավորությունները։ Մարդիկ կրում են կոնտակտային ոսպնյակներ բազմաթիվ', 'Մարդիկ կրում են կոնտակտային ոսպնյակներ բազմաթիվ պատճառներով։', 0.4915254237288136, 36695, 36760, 607]
    1108it [05:28,  4.85it/s][' հա՜, հա՜, հա՜…։Այն', 'Հա՛՜, հա՜, հա..', 0.5, 67528, 67547, 1107]
    1477it [07:12,  7.10it/s][' - Ո՞ւր ես գնում։', 'Ես արնր ես գնում։', 0.38461538461538464, 89741, 89758, 1477]
    1818it [08:52,  3.42it/s][' գիշեր։- նրա մեջ մ', '-Բարի լիշեր.', 0.8571428571428571, 110543, 110561, 1818]
    2632it [12:36,  2.28it/s][' էլի կասե՞ս սուտ', ' Էլիկ ասես Սուտա.', 0.38461538461538464, 159899, 159915, 2631]
    2935it [13:57,  5.08it/s][' Տգետ, պառավ, գինեվաճառ…։Նրա', 'Տկետ բառավ կին է վաճառը։', 0.4, 178049, 178077, 2934]
    3789it [17:57,  1.56it/s]['նկարները և այլն։ Ապրել են գաղութներով՝ տաք ծովերում և ոչ մեծ խորութ', 'Ապրել են գաղութներով, տաք ծովերում և ոչ մեծ խորություններում։', 0.4642857142857143, 229251, 229318, 3787]
    4281it [20:14,  3.53it/s]
    """
    print("Finished Matching the texts")
    benchmark = Benchmark(args.target_base, args.predictions_manifest)
    stats = benchmark.get_benchmark()
    benchmark.analyze_and_save_benchmark(stats, args.output_file)



import pandas as pd

from .factory import VAC


def run_pipeline(**kwargs) -> pd.DataFrame:
    return VAC.run_end2end(**kwargs)


if __name__ == '__main__':
    run_pipeline(manifest_file= r"D:\CapstoneThesisArmASR\vac_aligner\experiments\end2end\manifest.json", batch_size=64,
    asr_input_file=r"D:\CapstoneThesisArmASR\vac_aligner\vac_aligner\data\matching\test_with_predictions17_replaced.json",
    target_base=r"D:\CapstoneThesisArmASR\vac_aligner\experiments\end2end\texts", init_aligner_after_asr=True)


import os
from dataclasses import asdict
from typing import Optional, Tuple, Union, List, Dict

import pandas as pd
from loguru import logger

from .dto import ASRConfig, ALIGNERConfig
from vac_aligner.asr import ASR, ASR_MAPPING
from vac_aligner.matching import ALIGNER_MAPPING, BaseAlignerVAC, Benchmark


def _validate_aligner_info(asr_info: Dict) -> ALIGNERConfig:
    try:
        validated_asr_info = ALIGNERConfig(**asr_info)
        return validated_asr_info
    except TypeError as e:
        raise ValueError(f"Invalid ASR info provided: {e}")


def _validate_asr_info(asr_info: Dict) -> ASRConfig:
    try:
        validated_asr_info = ASRConfig(**asr_info)
        return validated_asr_info
    except TypeError as e:
        raise ValueError(f"Invalid ASR info provided: {e}")


class VAC:
    def __init__(self, language, asr_info: Optional[Dict], aligner_matching_info: Optional[Dict],
                 init_aligner_after_asr: bool = False):
        self.language = language
        self.asr_info = _validate_asr_info(asr_info) if asr_info else None
        self.aligner_matching_info = _validate_aligner_info(aligner_matching_info) if aligner_matching_info else None

        if isinstance(self.asr_info, ASRConfig):
            self.asr_model_name = self._get_asr_model_name(language)
            self.asr_model = ASR(self.asr_model_name, self.asr_info)

        if not init_aligner_after_asr:
            self.init_aligner()

        self.init_aligner_after_asr = init_aligner_after_asr

    def init_aligner(self):
        if isinstance(self.aligner_matching_info, dict):
            self.aligner_matching_info = ALIGNERConfig(**self.aligner_matching_info)
        self.aligner_class = self._get_aligner_class(self.language)
        self.aligner: BaseAlignerVAC = self.aligner_class.from_nemo_manifest(**asdict(self.aligner_matching_info))

    @staticmethod
    def _get_aligner_class(language):
        """Map language to corresponding Aligner Class"""
        aligner = ALIGNER_MAPPING.get(language)
        if not aligner:
            raise ValueError(f"Cannot retrieve aligner. Unsupported language: {language}")
        return aligner

    @staticmethod
    def _get_asr_model_name(language) -> str:
        model_name = ASR_MAPPING.get(language)
        if not model_name:
            raise ValueError(f"Cannot retrieve aligner. Unsupported language: {language}")
        return model_name

    def combine_transcript_and_align(self, predictions_manifest: str, save_manifest_path: str,
                                     output_file: str, target_base: str) -> List[Tuple[str, str, float, int, int]]:
        chunks, combined_transcript = self.aligner_class.combine_transcript(predictions_manifest, output_file)
        logger.info("Combined the MCV files into single transcript")
        self.aligner: BaseAlignerVAC = self.aligner_class(combined_transcript, chunks,
                                                          save_manifest_path, target_base=target_base)
        matches_sorted = self.aligner.align(0.35)
        return [tuple(x[:-1]) for x in matches_sorted]

    def run_asr(self, wav_files: Union[str, List[str]], save_dir: Optional[str] = None,
            save_manifest: Optional[str] = None, test_manifest: Optional[str] = None):

        chunk_data = self.asr_model.run(
            save_dir=save_dir,
            wav_files=wav_files,
            save_manifest=save_manifest,
            test_manifest=test_manifest
        )

        if self.init_aligner_after_asr:
            print("Initializing Aligner with the ASR manifest")
            self.init_aligner()

        return chunk_data

    def align_match(self, cer_bound: float = 0.35) -> List:
        return self.aligner.align(cer_bound)

    def __call__(self, wav_files: Union[str, List[str]], save_dir: Optional[str] = None,
            save_manifest: Optional[str] = None, test_manifest: Optional[str] = None):

        self.run_asr(
            save_dir=save_dir,
            wav_files=wav_files,
            save_manifest=save_manifest,
            test_manifest=test_manifest
        )

        self.aligner.align(0.35)

    @classmethod
    def run_end2end(cls, manifest_file: str, asr_input_file: str, language: str = "hy", init_aligner_after_asr = False,
                    target_base: Optional[str] = None, ending_punctuations: Optional[str] = '․,։', hf_token: str = '',
                    cer_bound: float=0.35,  batch_size: int = 15, gpu_id: Optional[int] = 0) -> pd.DataFrame:
        """Initialising VAC and running ASR -> obtained predictions manifest, run Aligner (combine transcript and match)

        Parameters:
            manifest_file : str
                The path to the NEMO manifest file (where to save the ASR predictions and then load for Aligner).
                Each line in this file should be a valid JSON string representing individual audio data with keys:
                    [audio_filepath, pred_text, duration]

            asr_input_file: str
                Either json manifest where audio_filepath indicates the audios to be inferenced on or a path to folder
                    containing .wav files. Used in ASR stage.

            language: str
                The language of the dataset. Currently, Default to "hy" (Armenian).

            target_base : Optional[str], optional
                The base path that could be prefixed to each audio file path in the manifest, indicating where to save
                    both ASR predictions (.txt s) and the matches (_matched.txt s). If not provided, will write the
                    text files in the same directory as the audio files are stored. Default is None.

            ending_punctuations : Optional[str]
                The punctuation symbols accumulated in single string (with most common sentence ending punctuation
                placed in the end of the string - such as: "...,;:.". Only necessary if transcript_path is None so that
                we combine the texts into a single transcript. Default is example for Armenian Language - '․,։'

            hf_token : str
                The hugging face token to fetch an ASR model. To get permissions, you must request access by the link
                    raise in the Permission Error when you first try to run. Each language will have it own ASR moddel.

            cer_bound : float
                The base CER threshold (upper bound) for matching algorithm. Defaults to 0.35

            gpu_id : (Optional[int])
                ID of the GPU to use (None if no GPU). Defaults to 0.

            batch_size : int
                The number of samples to process in a batch.
        """
        os.makedirs(target_base, exist_ok=True)
        os.makedirs(os.path.dirname(manifest_file), exist_ok=True)

        if not hf_token:
            hf_token = os.environ.get("HF_TOKEN")
            if hf_token is None:
                model_name = ASR_MAPPING[language]
                raise ValueError("Please, provide hf_token or directly set the hugging face token as ENV - HT_TOKEN."
                                 f"You might also need to request access at: https://huggingface.co/{model_name}")

        asr_config = {
            'gpu_id': gpu_id,
            'hf_token': hf_token,
            'batch_size': batch_size,

        }

        aligner_config = {
            'use_id': False,
            'transcript_path': None,
            'target_base': target_base,
            'manifest_file': manifest_file,
            'ending_punctuations': ending_punctuations,
            'save_manifest_path': manifest_file.replace(".json", "_vac.json")
        }

        vac = cls(language, asr_config, aligner_config, init_aligner_after_asr=init_aligner_after_asr)
        print("Initialized VAC!")
        asr_payload = dict(
            save_dir=target_base,
            save_manifest=manifest_file
        )
        if asr_input_file.endswith(".json"):
            asr_payload['wav_files'] = None
            asr_payload['test_manifest'] = asr_input_file
        else:
            asr_payload['wav_files'] = asr_input_file

        vac.run_asr(**asr_payload)
        print("Finished ASR inference!")

        vac.align_match(cer_bound)
        print("Finished Matching the texts. Now, performing the Benchmark")
        benchmark = Benchmark(target_base, manifest_file)
        stats = benchmark.get_benchmark()
        benchmark.analyze_and_save_benchmark(stats, os.path.join(target_base, "benchmark.csv"))
        return stats


if __name__ == '__main__':
    VAC.run_end2end(manifest_file=r"D:\CapstoneThesisArmASR\vac_aligner\experiments\end2end\manifest.json", batch_size=64,
                    asr_input_file=r"D:\CapstoneThesisArmASR\vac_aligner\vac_aligner\data\matching\test_with_predictions17_replaced.json",
                    target_base=r"D:\CapstoneThesisArmASR\vac_aligner\experiments\end2end\texts", init_aligner_after_asr=True)


from typing import Literal


import regex as re


def clean_corpus(text: str, source: Literal['prediction', 'original'] = 'original') -> str:
    """Corpus specific (source or predicted) text processing to better align based on CER threshold"""
    if source == "original":
        verjaket_origin = ":"
        verjaket_nemo = "։"
        replace_pairs = [("\n", " "), (" — ", " "), ("— ", " "), (" —", " "), (verjaket_origin, verjaket_nemo)]
        for pair in replace_pairs:
            text = text.replace(pair[0], pair[1])
        text = re.sub("\s+", " ", text)
        text = text.strip().replace(" և ", " եւ ")
    else:
        text = text.replace(" - ", " ").replace(" -", " ").replace("- ", " ")
        text = re.sub("\s+", " ", text)
    return text


def lowercase(token):
    """SPECIAL Name such as DAVID -> Special Name such as David"""
    if all([char.isupper() for char in token]):
        return token.lower().title()
    return token


"""
VAC - VAD-ASR-CER (Matching)

This is the VAC pipeline's last step - CER based matching!
==========================================================

A subpackage for both matching the chunked texts with single, long transcript and benchmarking (if labels are available)

"""

from .align import BaseAlignerVAC, ArmenianAlignerVAC
from .benchmark_on_mcv import Benchmark

ALIGNER_MAPPING = {
    "hy": ArmenianAlignerVAC
}


import os
import json
from abc import ABC, abstractmethod
from typing import List, Tuple, Optional


import regex as re
from tqdm import tqdm
from loguru import logger

from .cutting import cut_extra_tokens_from_match


data_item = Tuple[str, str, float]


class BaseAlignerVAC(ABC):
    def __init__(self, reference_text: str, chunks: List[Tuple], predictions_manifest_path: str,
                 target_base: Optional[str] = None, lost_search_cer: float = 0.5,
                 shift_back_minimum_bound: int = 8, shift_back_indicator: int = 4,
                 search_segment_length_upper_bound: int = 40, reduce_long_search_segment: int = 30):
        """
        Initializes the AlignerVAC* instance with necessary parameters to align audio chunks
        with a reference text using predicted ASR transcriptions.

        AlignerVAC* - operates within the VAC (VAD -> ASR -> CER-Matching) framework proposed in <>


        Parameters
        ----------
        reference_text : str
            The original long string to compare against, typically a combined transcript.
        chunks : List[Tuple[str, str, float]]
            A list of tuples in a format similar to NeMo, where each tuple contains:
                [audio_filepath, pred_text (ASR inference), audio_duration].
        predictions_manifest_path : str
            Path to a NeMo format manifest (json) where predictions are stored.
            Each file's row (dict) has the format:
                [audio_filepath, text, pred_text (ASR inference), audio_duration].
        target_base : Optional[str], optional
            The directory where matched texts will be stored as a single .txt per audio file.
            If not specified, defaults to a subdirectory 'chunk_texts_mcv' under predictions_manifest_path.
        lost_search_cer : float, optional
            Character Error Rate (CER) threshold indicating that the alignment is lost
            if sequentially exceeded. Default is 0.5.
        shift_back_minimum_bound : int, optional
            The minimum number of characters to backtrack when handling sequentially problematic chunks
            to resume search for the next chunk. Default is 8.
        shift_back_indicator : int, optional
            Number of consecutive problematic search runs required before initiating a shift back. Default is 4.
        search_segment_length_upper_bound : int, optional
            The maximum allowable search segment length, beyond which the cut_from_both_ends method
            might fail to handle extra tokens. Default is 40.
        reduce_long_search_segment : int, optional
            The number of characters to trim from the ends of a long search segment during processing. Default is 30.

        """
        self.target_base = target_base
        self.lost_search_cer = lost_search_cer
        self.shift_back_minimum_bound = shift_back_minimum_bound
        self.shift_back_indicator = shift_back_indicator
        self.search_segment_length_upper_bound = search_segment_length_upper_bound
        self.reduce_long_search_segment = reduce_long_search_segment

        self.memory = [chunks, reference_text, predictions_manifest_path]
        self.restore_memory()

    @property
    @abstractmethod
    def language_id(self) -> str:
        """Identifier for the language (e.g. Armenian, Georgian, etc.)"""
        ...

    def restore_memory(self):
        """Used to restore original Texts and Chunks to (re)run the algorithm"""
        self.chunks, self.reference_text, self.manifest_path = self.memory
        self.cut_the_header()

        self.durs = []
        self.matches = []

        self.shift_back = 0
        self.current_time = 0
        self.search_start_pos = 0  # Start searching from the beginning of the original text

        self.previous_pack = None
        self.try_combinations = False
        self.sequential_big_mismatches = []  # store metadata (chunks) of sequential mismatches (CER > 0.5)
        self.cer_cum = [0 for _ in range(15)]  # we terminate the algorithm if 15times in a row CER is 0.5+

    @property
    def default_ellipsis(self):
        """A punctuation harming the overall CER which can be ignored if ASR failed to predict them"""
        return "..."

    @default_ellipsis.setter
    def default_ellipsis(self, value):
        """
        Sets the default ellipsis value with the provided string.
        Ensures that the input is a string and exactly three periods (language specific), else raises a ValueError.

        Parameters:
        value (str): A string to set as the new default ellipsis.
        """
        if isinstance(value, str) and len(value) == 3:
            self._default_ellipsis = value
        else:
            raise ValueError("The default ellipsis must be a string containing exactly three periods.")

    def cut_the_header(self):
        if len(self.reference_text.split(" ")[0]) == 1 or len(self.reference_text.split("\n")[0]) == 1:
            self.reference_text = self.reference_text[2:]
            logger.info(f'Cut the Header: {self.chunks[0]}')
            self.chunks = self.chunks[1:]

    def get_target_path(self, original_path: str) -> str:
        if "\\" in original_path:
            audio_name = original_path.split("\\")[-1]
        else:
            audio_name = original_path.split("/")[-1]

        if self.target_base:
            target_path = os.path.join(self.target_base, audio_name.replace(".wav", "_matched.txt"))
        else:
            target_path = os.path.join(os.path.dirname(self.manifest_path), "chunk_texts_mcv",
                                       audio_name.replace(".wav", "_matched.txt"))
        return target_path

    def dump_match_info(self, target_path: str, best_match_text: str, match_info: dict):
        with open(target_path, "w", encoding='utf-8') as f:
            f.write(best_match_text.strip())

        if not os.path.exists(self.manifest_path):
            os.makedirs(os.path.dirname(self.manifest_path), exist_ok=True)
            with open(self.manifest_path, "w", encoding='utf-8') as f:
                f.write(json.dumps(match_info, ensure_ascii=False) + "\n")
        else:
            with open(self.manifest_path, "a", encoding='utf-8') as f:
                f.write(json.dumps(match_info, ensure_ascii=False) + "\n")

    def handle_mismatches(self) -> Tuple[Tuple[int, int], str, str, bool]:
        """Handle sequentiall big mismatches by coming to the past, and taking the first best output"""
        sequential_big_mismatches = sorted(self.sequential_big_mismatches, key=lambda x: x[2])
        window_text, chunk_text, cer, start_pos, end_pos, _ = sequential_big_mismatches[0]
        best_cer = cer
        best_match_range = (start_pos, end_pos)
        best_match_text = window_text
        found_acceptable_match = True  # after handling this problematic chunk need to proceed to the next chunk
        self.shift_back = min(len(window_text) // 2, self.shift_back_minimum_bound )
        tokens_ = re.findall(r'\S+|\s+', window_text)
        if not tokens_:
            logger.warning(sequential_big_mismatches)
            raise ValueError

        if len(tokens_[-1]) <= 2:  # punctuation
            if len(" ".join(tokens_[-2:])) > self.shift_back:  # get back from whole token, not in the middle
                self.shift_back = len(" ".join(tokens_[-2:])) + 1
        else:
            if len(" ".join(tokens_[-1])) > self.shift_back:  # get back from whole token, not in the middle
                self.shift_back = len(" ".join(tokens_[-1])) + 1

        self.search_start_pos = end_pos - self.shift_back
        self.previous_pack = [len(window_text), best_cer]
        self.try_combinations = True
        logger.warning(f"Potentially problematic: {sequential_big_mismatches[0]}")
        return best_match_range, chunk_text, best_match_text, found_acceptable_match


    @property
    @abstractmethod
    def ending_punctuations(self) -> str:
        """Language Specific sentence ending punctuations such as ։․,"""
        pass

    @staticmethod
    @abstractmethod
    def language_specific_cleaning(chunk: str):
        """Language Specific text cleaning should be implemented here"""
        return chunk

    @abstractmethod
    def language_specific_postprocessing(self, best_match_text: str, best_match_range: Tuple[int, int]) -> Tuple[str, Tuple[int, int]]:
        """Based on the language, some specific grammatical rules might be needed"""
        return best_match_text, best_match_range

    def update_window(self, window_text: str, chunk_text: str, start_pos: int):
        """Need to Update the Rolling Window over Source Transcript (if it is >> chunk_text)"""
        new_window_text, cer, to_shift = cut_extra_tokens_from_match(window_text,
                                                                     chunk_text,
                                                                     self.try_combinations)

        clen = len(chunk_text)
        wlen = len(window_text)
        if (clen * 6 < len(window_text)) or (clen < self.search_segment_length_upper_bound < wlen):
            smaller_window, cer2, shift2 = cut_extra_tokens_from_match(window_text[-self.reduce_long_search_segment:],
                                                                       chunk_text,
                                                                       self.try_combinations)
            if cer2 <= cer:
                start_pos += wlen - self.reduce_long_search_segment  # adjust starting position as cut last 30chars
                new_window_text, cer, to_shift = smaller_window, cer2, shift2

        return new_window_text, cer, to_shift, start_pos

    def clean_chunk_text(self, chunk_text: str) -> str:
        chunk_text = chunk_text.replace(" - ", " ").replace(" -", " ").replace("- ", " ")
        chunk_text = self.language_specific_cleaning(chunk_text)
        chunk_text = re.sub("\s+", " ", chunk_text)
        return chunk_text

    def check_lost_scenario(self):
        if all([x > self.lost_search_cer for x in self.cer_cum]):
            logger.warning("transcripts got lost the matching, no point to continue")
            return self.matches

    def parse_chunk_and_prepare_iter(self, chunk: data_item) -> Tuple[Tuple, Tuple]:
        original_path, chunk_text, chunk_duration = chunk
        chunk_text = self.clean_chunk_text(chunk_text)
        target_path = self.get_target_path(original_path)
        os.makedirs(os.path.dirname(target_path), exist_ok=True)

        best_cer = float('inf')
        best_match_range = (0, 0)
        best_match_text = ""
        found_acceptable_match = False
        self.sequential_big_mismatches = []

        if self.previous_pack:  # previous chunk was added if len(sequential_big_mismatches) > 4:
            if self.previous_pack[-1] > 0.3:
                self.shift_back += self.previous_pack[0]
                self.search_start_pos -= self.previous_pack[0]
                self.search_start_pos = max(self.search_start_pos, 0)
                self.try_combinations = True

        return ((original_path, target_path, chunk_text, chunk_duration),
                (best_cer, best_match_range, best_match_text, found_acceptable_match))

    def align(self, cer_threshold=0.35):
        try:
            for ind, chunk in tqdm(enumerate(self.chunks)):
                self.check_lost_scenario()

                chunk_metadata, init_states = self.parse_chunk_and_prepare_iter(chunk)
                original_path, target_path, chunk_text, chunk_duration = chunk_metadata
                best_cer, best_match_range, best_match_text, found_acceptable_match = init_states

                for start_pos in range(self.search_start_pos, len(self.reference_text)):
                    if len(self.sequential_big_mismatches) > self.shift_back_indicator:
                        best_match_range, chunk_text, best_match_text, found_acceptable_match = self.handle_mismatches()
                        break

                    to_add = 0
                    end_pos = start_pos + len(chunk_text)
                    window_text = self.reference_text[start_pos:end_pos + 6 + self.shift_back]
                    if self.default_ellipsis in window_text and self.default_ellipsis not in chunk_text:
                        window_text = window_text.replace(self.default_ellipsis, "")
                        to_add = 3  # as ignored 3chars from source segment

                    new_window_text, cer, to_shift, start_pos = self.update_window(window_text, chunk_text, start_pos)
                    start_pos += to_shift
                    window_text = new_window_text

                    # end_pos - (len(window_text) - len(new_window_text))
                    end_pos = start_pos + len(window_text) + to_add

                    self.sequential_big_mismatches.append([new_window_text, chunk_text, cer, start_pos, end_pos, ind - 1])
                    self.try_combinations = False
                    self.shift_back = 0

                    if cer < best_cer:
                        best_cer = cer
                        best_match_range = (start_pos, end_pos)
                        best_match_text = window_text

                    if cer <= cer_threshold:  # Break if CER is below the threshold, indicating a good enough match
                        self.previous_pack = None
                        found_acceptable_match = True
                        self.search_start_pos = end_pos
                        break

                if not found_acceptable_match:
                    self.search_start_pos += len(chunk_text)  # adjust search start position conservatively

                best_match_text, best_match_range = self.language_specific_postprocessing(best_match_text,
                                                                                          best_match_range)

                match_info = {
                    'audio_filepath': original_path,
                    'matched_text_path': target_path,
                    'text': best_match_text.strip(),
                    'chunk_text': chunk_text,
                    'start_time': self.current_time,
                    'end_time': self.current_time + chunk_duration,
                    'best_match_range': best_match_range,
                    'cer': best_cer,
                    'duration': chunk_duration,
                }

                self.cer_cum = self.cer_cum[1:] + [best_cer]
                self.matches.append(match_info)
                self.durs.append(len(self.sequential_big_mismatches))
                self.current_time += chunk_duration
                self.dump_match_info(target_path, best_match_text, match_info)

        except KeyboardInterrupt:
            logger.error(self.matches)

        return self.matches

    @classmethod
    def from_nemo_manifest(cls, manifest_file: str,  save_manifest_path: str, transcript_path: Optional[str] = None,
                           use_id: bool = False, target_base: Optional[str] = None, ending_punctuations: str = '․,։'):
        """
        Create an instance by reading and processing a NEMO manifest and a transcript file.

        Parameters:
            manifest_file : str
                The path to the NEMO manifest file. Each line in this file should be a valid JSON string representing
                    individual audio data with keys [audio_filepath, pred_text, duration]

            transcript_path : Optional[str]
                The path to the file containing the combined transcript of all audio files.
                If None, will calculate (combine) based on the "text" key in the manifest_file.

            save_manifest_path : str
                The path where the processed manifest data will be saved or utilized.

            use_id : bool, optional
                Determines whether to sort the audio chunks based on their ID. Useful when IDs are used to denote the
                    order of audio files/chunks. If provided, then manifest_file should have id field. Default is False.

            target_base : Optional[str], optional
                The base path that could be prefixed to each audio file path in the manifest, useful for cases where
                    the manifest paths are relative. Default is None.

            ending_punctuations : Optional[str]
                The punctuation symbols accumulated in single string (with most common sentence ending punctuation
                placed in the end of the string - such as: "...,;:.". Only necessary if transcript_path is None so that
                we combine the texts into a single transcript. Default is example for Armenian Language - '․,։'

        Returns:
            An instance of the class populated with the combined transcript, sorted (if specified) and filtered chunks

        Example:
            # Create a new instance using a manifest file, a transcript, and specify the use of IDs
            instance = ClassName.from_nemo_manifest('path/to/manifest.json', 'path/to/transcript.txt',
                                                    'path/to/save/manifest', target_base="abs_path_cwc", use_id=True)
        """
        if transcript_path is None:
            combined_transcript = ""
        else:
            with open(transcript_path, "r", encoding='utf-8') as f:
                combined_transcript = f.read()

        chunks = []
        with open(manifest_file, "r", encoding='utf-8') as f:
            for line in f:
                item = json.loads(line)
                chunk = (item["audio_filepath"], item["pred_text"], item["duration"], item.get("id"))
                chunks.append(tuple(chunk))
                if transcript_path is None:
                    text = item["text"]
                    if text[-1] not in ":,." + ending_punctuations:
                        text = text + (ending_punctuations[-1] or "։")
                    else:
                        text = text + " "
                    combined_transcript += text

        if use_id:
            chunks = sorted(chunks, key=lambda x: x[-1])
        chunks = [tuple(chunk[:-1]) for chunk in chunks]
        return cls(combined_transcript, chunks, save_manifest_path, target_base=target_base)

    @staticmethod
    def combine_transcript(predictions_manifest: str, output_file: str, ending_punctuations: str) \
            -> Tuple[List[data_item], str]:
        chunks = []
        combined_transcript = ""
        with open(fr"{predictions_manifest}", "r", encoding='utf-8') as f:
            for line in f:
                item = json.loads(line)
                chunks.append((item["audio_filepath"], item["pred_text"], item["duration"]))
                text = item["text"]
                if text[-1] not in ":,." + ending_punctuations:
                    text = text + (ending_punctuations[-1] or "։")
                else:
                    text = text + " "
                combined_transcript += text

        with open(fr"{output_file}", "w", encoding='utf-8') as t:
            t.write(combined_transcript)

        return chunks, combined_transcript


class ArmenianAlignerVAC(BaseAlignerVAC):
    """Aligner for Armenian Language"""

    @property
    def language_id(self):
        return "hy"

    @property
    def ending_punctuations(self) -> str:
        return '․,։'

    @staticmethod
    def language_specific_cleaning(chunk_text: str):
        """Language Specific text cleaning should be implemented here"""
        return chunk_text.replace("Ե ՛վ", "Եվ")

    def language_specific_postprocessing(self, best_match_text: str, best_match_range: Tuple[int, int]) -> Tuple[str, Tuple[int, int]]:
        """Based on the language, some specific grammatical rules might be needed"""
        if best_match_text.startswith("ւ"):  # specific case for Armenian letter "ու"
            if self.reference_text[best_match_range[0] - 1] == "ո":
                best_match_text = "ո" + best_match_text
                best_match_range = (best_match_range[0] - 1, best_match_range[1])

        return best_match_text, best_match_range


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Process the predictions manifest and output a combined transcript.")
    parser.add_argument("predictions_manifest", type=str, help="Path to the manifest file storing predictions.",
                        default=r"D:\CapstoneThesisArmASR\data\cv-corpus-17.0-2024-03-15\hy-AM\test\vac_mcv_manifest.json")
    parser.add_argument("combined_transcript", type=str, help="Combined MCV transcript.")
    args = parser.parse_args()

    # matches_sorted = ArmenianAlignerVAC(args.combined_transcript, chunks, args.predictions_manifest).align(0.35)



#!/usr/bin/env python

import pytest
from vac_aligner.matching.metrics import calculate_cer, calculate_wer, calculate_cer_full


def test_calculate_cer_completely_different():
    "Assuming completely different strings of the same length"
    assert calculate_cer_full("hello", "abcde") == 1, "CER for completely different strings should be 1"


def test_calculate_cer_substitution():
    "One substitution"
    assert 1 > calculate_cer("hello", "jello") > 0, "CER should be between 0 and 1 for one substitution"


def test_calculate_cer_identical():
    assert calculate_cer("hello", "hello") == 0, "CER for identical strings should be 0"


def test_calculate_cer_insertion():
    assert 1 > calculate_cer("hello", "helloo") > 0, "CER should be between 0 and 1 for one insertion"


def test_calculate_cer_deletion():
    assert 1 > calculate_cer("hello", "hell") > 0, "CER should be between 0 and 1 for one deletion"


def test_calculate_cer_empty_strings():
    assert calculate_cer("", "") == 0, "CER for two empty strings should be 0"
    assert calculate_cer("hello", "") == 1, "CER should be 1 when comparing any string to an empty string"
    assert calculate_cer("", "hello") == 1, "CER should be 1 when comparing empty string to any string"

# Assuming calculate_cer_full is implemented correctly
# Here are basic tests for it, mirroring the tests for calculate_cer but potentially with different expectations


def test_calculate_cer_full_identical():
    assert calculate_cer_full("hello", "hello") == 0, "Full CER for identical strings should be 0"


def test_calculate_wer_completely_different():
    """Assuming completely different strings with the same number of words"""
    assert calculate_wer("hello world", "goodbye there") == 1, "WER for completely different strings should be 1"


def test_calculate_wer_substitution():
    """One substitution in a two-word sentence"""
    assert calculate_wer("hello world", "hello there") == 0.5, "WER should be 0.5 for one substitution out of two words"


def test_calculate_wer_insertion():
    """One insertion"""
    assert calculate_wer("hello world", "hello big world") == 1/2, "WER should be 1/2 for one insertion in a three-word sentence"


def test_calculate_wer_deletion():
    """One deletion"""
    assert calculate_wer("hello big world", "hello world") == 1/3, "WER should be 1/3 for one deletion in a three-word sentence"


def test_calculate_wer_empty_strings():
    assert calculate_wer("", "") == 0, "WER for two empty strings should be 0"
    assert calculate_wer("hello world", "") == 1, "WER should be 1 when comparing any string to an empty string"
    assert calculate_wer("", "hello world") == 1, "WER should be 1 when comparing empty string to any string"


def test_calculate_wer_multiple_errors():
    "Multiple errors including substitution, insertion, and deletion"
    original = "hello world here"
    modified = "hi world there"
    # Expected WER: (1 substitution: hello->hi, 1 substitution: here->there) / 3 words = 2/3
    assert calculate_wer(original, modified) == 2/3, "WER should account for multiple types of errors"



import os
import json
import wave
import contextlib
from typing import List, Tuple, Union, Optional

from vac_aligner.dto import ASRConfig, ChunkData

from glob import glob


class ASR:
    """
    Class which  perform ASR inference in a loop and writes to NeMo manifest.
    NeMo and torch related imports are made local as we have `pip install vac_aligner['full']` if asr is needed
    """
    def __init__(self, model_name: str, asr_config: ASRConfig, skip_existing: bool = True):
        from nemo.collections.asr.models import EncDecCTCModelBPE

        self.asr_config = asr_config
        self.skip_existing = skip_existing
        self.asr_model_bpe: EncDecCTCModelBPE = self.load_model(model_name)

    def load_model(self, model_name: str):
        """Load the model (currently support only HF) and map to correct device"""
        import torch
        from nemo.collections.asr.models import ASRModel

        if not os.environ.get("DISABLE_HF", False):
            model = ASRModel.from_pretrained(model_name)
            if self.asr_config.gpu_id is None:
                device = torch.device('cpu')
            else:
                device = torch.device(f'cuda:{self.asr_config.gpu_id}' if torch.cuda.is_available() else 'cpu')
            model = model.to(device)
            return model
        else:
            raise NotImplemented

    @staticmethod
    def extract_wav_files(wav_files: Union[str, List[str]], test_manifest: Optional[str] = None) -> Tuple[List[str], List[Optional[str]]]:
        if test_manifest:
            texts = []
            wav_files = []
            with open(test_manifest, "r", encoding="utf-8") as f:
                for line in f:
                    item = json.loads(line)
                    texts.append(item.get("text", None))
                    wav_files.append(item["audio_filepath"])
            print("Loaded only Test audios given in the optional manifest!")
        else:
            if isinstance(wav_files, str):
                wav_files = glob(wav_files+"/*.wav")
            texts = [None for _ in range(len(wav_files))]
        return wav_files, texts

    def run(self, wav_files: Union[str, List[str]], save_dir: Optional[str] = None,
            save_manifest: Optional[str] = None, test_manifest: Optional[str] = None) -> List[ChunkData]:
        """

        Inference with while (batched) as loading too many audios decreases the performance

        Parameters:
            wav_files : Union[str, List[str]]
                If str then have to be the path to the directory where the audios (.wav) where stored or a list
                    containing the paths (audio_filepath)

            save_manifest : str
                The path to the nemo manifest file to save the chunk info.

            save_dir : Optional[str]
                The path to the folder where to ASR predictions (single .txt per audio_filepath). If not provided,
                    will store in the same place as wav_files are stored. Defaults to None.

            test_manifest : Optional[str]
                If there is a nemo manifest available, then it can also be provided to replace wav_files.
                The format of manifest should just contain: [audio_filepath]. If not provided, will only refer to
                    wav_files. Default is None.

        Returns:
             A list of chunks with corresponding info (audio_filepath, chunk_text, chunk_text_path, audio_duration, ...)
        """
        wav_files, texts = self.extract_wav_files(wav_files, test_manifest)
        wav2text = {wav_files[i]: texts[i] for i in range(len(wav_files))}
        print(f"#of Audios to Process! {len(wav_files)}")
        b_size = self.asr_config.batch_size
        batch_end = b_size * 1
        full_preds = []

        while batch_end <= len(wav_files):
            batch = wav_files[batch_end - b_size:batch_end]
            if self.skip_existing and all([os.path.exists(self.wav2text_path(x, save_dir)) for x in batch]):
                batch_end += b_size
                continue
            preds = self.asr_model_bpe.transcribe(batch)
            for i, wav in enumerate(batch):
                duration = self.get_wav_duration(wav)
                chunk_txt = self.wav2text_path(wav, save_dir)
                # if os.path.exists(chunk_txt):
                #     continue
                with open(chunk_txt, "w", encoding='utf-8') as f:
                    f.write(preds[i])
                full_preds.append(
                    ChunkData(
                        audio_path=wav,
                        duration=duration,
                        save_path=chunk_txt,
                        asr_prediction=preds[i],
                        text=wav2text[wav]
                    )
                )
            batch_end += b_size

            if batch_end % 80 == 0:
                print(chunk_txt)
                print(f"Batch reached: {batch_end}")

        not_processed = list(filter(lambda x: not os.path.exists(self.wav2text_path(x, save_dir)), wav_files))
        if not_processed:
            print("Running on the last (partial) batch!")
            text_batch = texts[batch_end - b_size:batch_end]
            preds = self.asr_model_bpe.transcribe(not_processed)
            for i, wav in enumerate(not_processed):
                chunk_txt = wav.replace(".wav", ".txt")
                if os.path.exists(chunk_txt):
                    continue
                with open(chunk_txt, "w", encoding='utf-8') as f:
                    f.write(preds[i])
                duration = self.get_wav_duration(wav)
                full_preds.append(
                    ChunkData(
                        audio_path=wav,
                        duration=duration,
                        save_path=chunk_txt,
                        asr_prediction=preds[i],
                        text=wav2text[wav]
                    )
                )

        print(f"#of Audios Successfully Processed! [{len(full_preds)}/{len(wav_files)}]")
        if save_manifest:
            self.write_to_manifest(full_preds, save_manifest)
        return full_preds

    @staticmethod
    def write_to_manifest(items: List[ChunkData], save_manifest: str):
        """Saving chunks into NeMo manifest with corresponding format (each raw is a json representing one chunk)"""
        with open(save_manifest, "w", encoding='utf-8') as f:
            for item in items:
                raw = {
                    'audio_filepath': item.audio_path,
                    'pred_text': item.asr_prediction,
                    'duration': item.duration,
                }
                text = item.text
                if text:
                    raw['text'] = text
                f.write(json.dumps(raw, ensure_ascii=False)+"\n")

    @staticmethod
    def wav2text_path(wav_filepath: str, save_dir: Optional[str] = None):
        if save_dir is None:
            return wav_filepath.replace(".wav", ".txt")
        audio_name = wav_filepath.split("\\")[-1].split("/")[-1]
        txt_filename = audio_name.replace(".wav", ".txt")
        txt_filepath = os.path.join(save_dir, txt_filename)
        return txt_filepath

    @staticmethod
    def get_wav_duration(wav_path: str) -> float:
        with contextlib.closing(wave.open(wav_path, 'rb')) as f:
            frames = f.getnframes()
            rate = f.getframerate()
            duration = frames / float(rate)
            return duration


if __name__ == "__main__":
    model_name = "Yeroyan/stt_arm_conformer_ctc_large"
    asr_config = ASRConfig(hf_token=os.environ['HF_TOKEN'], batch_size=24)
    asr = ASR(model_name, asr_config)
    asr.run(
        save_dir=r"D:\CapstoneThesisArmASR\vac_aligner\experiments\asr_preds",
        save_manifest=r"D:\CapstoneThesisArmASR\vac_aligner\experiments\asr.json",
        wav_files=r"D:\Capstone_Thesis\cv-corpus-17.0-2024-03-15\hy-AM\resampled_clips",
        test_manifest=r"D:\CapstoneThesisArmASR\vac_aligner\vac_aligner\data\matching\test_with_predictions17_replaced.json"
    )


#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()


def read_requirements():
    with open('requirements.txt') as req:
        return req.read().strip().split('\n')


extras_require = {
    'full': [
        'nemo_toolkit[asr]'
    ]
}

test_requirements = ['pytest>=3', ]

setup(
    author="Ara Yeroyan",
    author_email='ar23yeroyan@gmail.com',
    python_requires='>=3.10',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    description="Aligning Very Long audio and text pairs through VAC pipeline",
    install_requires=read_requirements(),
    extras_require=extras_require,
    license="Apache Software License 2.0",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='vac_aligner',
    name='vac_aligner',
    packages=find_packages(include=['vac_aligner', 'vac_aligner.*']),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/Ara-Yeroyan/vac_aligner',
    download_url='https://github.com/Ara-Yeroyan/vac_aligner/archive/refs/tags/v0.0.2.tar.gz',
    version='0.0.4',
    zip_safe=False,
)



"""
VAC - VAD-ASR-CER (Matching) Pipeline package

1. VAD: Voice Activity Detection
2. ASR: Automatic Speech Recognition
3. CER: Character Error Rate

**Usage***

VAD splits long audio into many multi-second chunks
ASR runs inference and gets predictions per chunk (chunk texts)
CER-Matching uses our algorithm to match the predicted text (with errors) to ground truth transcript

**Results**

Having multi hour long audio and texts, you can obtain many multi-second audio chunks and corresponding texts (98% acc)

"""

__author__ = """Ara Yeroyan"""
__email__ = 'ar23yeroyan@gmail.com'
__version__ = '0.0.4'

from .vac_pipeline import run_pipeline